#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import abc
from dataclasses import dataclass

from entity_read.sql.context import SqlContext

SQL = str


@dataclass(kw_only=True, frozen=True, slots=True)
class Node(abc.ABC):
    # TODO: replace key_to_expression with "existent attributes", where present table columns and relations
    @abc.abstractmethod
    def to_sql(self, context: SqlContext, key_to_expression: dict = None) -> SQL:
        ...

    @abc.abstractmethod
    def requires(self) -> list:
        ...



#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import typing
from dataclasses import dataclass

# from entity_read.select.atoms.atoms import NodeType
from .abstract import Selectable
from .column import Column


@dataclass(kw_only=True, frozen=True, slots=True)
class Literal(Selectable):
    # type = NodeType.literal
    value: typing.Any
    # alias: str = None

    def to_sql(self, context, key_to_expression = None) -> str:
        if not context.bind_params:
            return context.register_parameter(self.value)

        value = list(self.value) if isinstance(self.value, set) else self.value
        result = str(value)
        if isinstance(value, list):
            if self.value:
                result = 'ARRAY' + result
            else:
                result = "'{}'"
        return result

    def requires(self) -> list['Column']:
        return []

    def get_name(self) -> str:
        return str(self.value)

    # def get_name(self) -> str:
    #     return self.alias


# @dataclass(kw_only=True, frozen=True, slots=True)
# class EStringLiteral(ELiteral):
#     value: str
#
#
# @dataclass(kw_only=True, frozen=True, slots=True)
# class EIntegerLiteral(ELiteral):
#     value: int



#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from typing import Any


class SqlContext:
    def __init__(self, self_table_name: str, key_to_param: dict | None = None):
        self.self_table_name = self_table_name
        self.bind_params = key_to_param is None
        self.key_to_param = key_to_param

    def get_self(self) -> str:
        return self.self_table_name

    def register_parameter(self, value: Any) -> str:
        assert self.key_to_param is not None, f'Attempt to register parameter({value}) with no key_to_param storage'
        if any(
                value == (param_val := (target_item := item)[1]) and type(value) == type(param_val)
                for item in self.key_to_param.items()
        ):
            key = target_item[0]
        else:
            key = f'l{len(self.key_to_param)}'
            self.key_to_param[key] = value
        return ":" + key



#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import dataclasses
from typing import Any, Callable, Awaitable

from no_value import NoValue
from sqlalchemy import ForeignKey, Column
from sqlalchemy.orm import relationship
from sqlalchemy.sql.type_api import TypeEngine

from .utils.defs import SA_DATACLASS_METADATA_KEY
from .utils.no_default import NoDefault


def entity_field(
        type_: type[TypeEngine] | TypeEngine | ForeignKey,
        foreign_key: ForeignKey = None,
        *,
        primary_key: bool = False, nullable: bool = True, unique: bool = False, index: bool = False,
        default: Any = None, server_default: Any = None, default_factory: Callable = None,
        insert_default: Callable[[...], Awaitable[Any]] = None, autoincrement: bool = None,
        identifier: bool = False
) -> dataclasses.field:
    if primary_key:
        nullable = True

    column_args = [type_, foreign_key]
    column_kwargs = {"primary_key": primary_key} if primary_key else {"nullable": nullable, "unique": unique}

    if default is not None:
        column_kwargs['default'] = default
    if server_default is not None:
        column_kwargs['server_default'] = server_default
    if index:
        column_kwargs['index'] = index
    if autoincrement is not None:
        column_kwargs['autoincrement'] = autoincrement

    metadata = {SA_DATACLASS_METADATA_KEY: Column(*column_args, **column_kwargs)}

    # extension attributes
    if insert_default:
        # TODO: check no "default" passed, think about using insert_default with None
        metadata['insert_default'] = insert_default
    metadata['identifier'] = identifier
    # metadata['descend'] = None

    kwargs: dict[str, Any] = {"metadata": metadata}
    if default_factory:
        kwargs['default_factory'] = default_factory
    else:
        # noinspection PyTypedDict
        kwargs['default'] = NoDefault
        if default:
            kwargs['default'] = default
        elif nullable:
            kwargs['default'] = None
        elif insert_default is not None or server_default is not None:
            kwargs['default'] = NoValue

    field = dataclasses.field(**kwargs)
    return field


def entity_relation_field(remote_class, *, init: bool = True, use_list: bool = True,
                          order_by: str | bool | list = False, lazy: str = "noload",
                          **relationship_kwargs) -> dataclasses.field:
    metadata = {
        SA_DATACLASS_METADATA_KEY: relationship(
            remote_class, uselist=use_list, order_by=order_by, lazy=lazy, **relationship_kwargs
        )}
    return dataclasses.field(metadata=metadata, init=init, default_factory=list)


def get_sqlalchemy_metadata(field: dataclasses.field) -> Any:
    return field.metadata[SA_DATACLASS_METADATA_KEY]
